@@ -1,6 +1,6 @@
 /***********************************************************************
  *
- * Copyright (C) 2011 Graeme Gott <graeme@gottcode.org>
+ * Copyright (C) 2011, 2012 Graeme Gott <graeme@gottcode.org>
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -21,6 +21,7 @@
 
 #include "appearance_dialog.h"
 
+#include <QCoreApplication>
 #include <QFile>
 #include <QGLShaderProgram>
 #include <QSettings>
@@ -76,6 +77,21 @@ static PFNGLDELETEBUFFERSPROC deleteBuffers = 0;
 typedef void (APIENTRYP PFNGLGENBUFFERSPROC) (GLsizei n, GLuint *buffers);
 static PFNGLGENBUFFERSPROC genBuffers = 0;
 
+// Shaders extension
+#ifndef GL_VERSION_2_0
+#define GL_SHADING_LANGUAGE_VERSION 0x8B8C
+#endif
+
+static QString glsl_version;
+
+// OpenGL 3 string
+#ifndef GL_VERSION_3_0
+#define GL_NUM_EXTENSIONS 0x821D
+#endif
+
+typedef const GLubyte* (APIENTRYP PFNGLGETSTRINGIPROC)(GLenum name, GLuint index);
+static PFNGLGETSTRINGIPROC getStringi = 0;
+
 // Vertex attribute object extension
 static GLuint vao_id = 0;
 
@@ -131,14 +147,13 @@ static inline void convertMatrix(const qreal* in, GLfloat* out)
 void GraphicsLayer::init()
 {
 	unsigned int state = 0;
-	int detected = 11;
+	unsigned int symbols = 0;
 
 	// Try to load multi-texture functions
 	activeTexture = (PFNGLACTIVETEXTUREPROC) getProcAddress("glActiveTexture");
 	clientActiveTexture = (PFNGLCLIENTACTIVETEXTUREPROC) getProcAddress("glClientActiveTexture");
-	if ((activeTexture != 0) && (clientActiveTexture != 0)) {
-		state |= MultiTextureFlag;
-		detected = 13;
+	if (activeTexture && clientActiveTexture) {
+		symbols |= MultiTextureFlag;
 	}
 
 	// Try to load vertex buffer object functions
@@ -147,49 +162,87 @@ void GraphicsLayer::init()
 	bufferSubData = (PFNGLBUFFERSUBDATAPROC) getProcAddress("glBufferSubData");
 	deleteBuffers = (PFNGLDELETEBUFFERSPROC) getProcAddress("glDeleteBuffers");
 	genBuffers = (PFNGLGENBUFFERSPROC) getProcAddress("glGenBuffers");
-	if ((bindBuffer != 0) && (bufferData != 0) && (bufferSubData != 0) && (deleteBuffers != 0) && (genBuffers != 0)) {
-		state |= VertexBufferObjectFlag;
-		if (state == Version15) {
-			detected = 15;
-		}
+	if (bindBuffer && bufferData && bufferSubData && deleteBuffers && genBuffers) {
+		symbols |= VertexBufferObjectFlag;
 	}
 
 	// Check for minimum supported programmable pipeline
-	if (QGLFormat::openGLVersionFlags() & QGLFormat::OpenGL_Version_2_1) {
+	QByteArray glsl = reinterpret_cast<const char*>(glGetString(GL_SHADING_LANGUAGE_VERSION));
+	int index = glsl.indexOf(' ');
+	if (index != -1) {
+		glsl.truncate(index);
+	}
+	glsl.replace('.', "");
+	if (QGLShaderProgram::hasOpenGLShaderPrograms() && (glsl >= "120")) {
 		state |= FragmentShadersFlag;
-		if (activeTexture != 0) {
-			state |= MultiTextureFlag;
+		symbols |= FragmentShadersFlag;
+		if (activeTexture) {
+			symbols |= MultiTextureFlag;
 		}
 	}
 
 	// Try to load vertex array object functions
-	if (state == Version21) {
-		detected = 21;
-		bindVertexArray = (PFNGLBINDVERTEXARRAYPROC) getProcAddress("glBindVertexArray");
-		deleteVertexArrays = (PFNGLDELETEVERTEXARRAYSPROC) getProcAddress("glDeleteVertexArrays");
-		genVertexArrays = (PFNGLGENVERTEXARRAYSPROC) getProcAddress("glGenVertexArrays");
-		if ((bindVertexArray != 0) && (deleteVertexArrays != 0) && (genVertexArrays != 0)) {
-			state |= VertexArrayObjectFlag;
-			detected = 30;
+	bindVertexArray = (PFNGLBINDVERTEXARRAYPROC) getProcAddress("glBindVertexArray");
+	deleteVertexArrays = (PFNGLDELETEVERTEXARRAYSPROC) getProcAddress("glDeleteVertexArrays");
+	genVertexArrays = (PFNGLGENVERTEXARRAYSPROC) getProcAddress("glGenVertexArrays");
+	if (bindVertexArray && deleteVertexArrays && genVertexArrays) {
+		symbols |= VertexArrayObjectFlag;
+	}
+
+	// Try to determine OpenGL extensions
+	QList<QByteArray> extensions;
+	const char* ext = reinterpret_cast<const char*>(glGetString(GL_EXTENSIONS));
+	if (!ext) {
+		getStringi = (PFNGLGETSTRINGIPROC) getProcAddress("glGetStringi");
+		if (getStringi) {
+			int count = 0;
+			glGetIntegerv(GL_NUM_EXTENSIONS, &count);
+			for (int i = 0; i < count; ++i) {
+				extensions += reinterpret_cast<const char*>(getStringi(GL_EXTENSIONS, i));
+			}
 		}
+	} else {
+		extensions = QByteArray(ext).split(' ');
+	}
+	foreach (const QByteArray& extension, extensions) {
+		if (extension == "GL_ARB_multitexture") {
+			state |= MultiTextureFlag;
+		} else if (extension == "GL_ARB_vertex_buffer_object") {
+			state |= VertexBufferObjectFlag;
+		} else if (extension == "GL_ARB_vertex_array_object") {
+			state |= VertexArrayObjectFlag;
+		}
+	}
+
+	// Find maximum supported state
+	int detected = 11;
+	switch (qMin(state, symbols)) {
+	case Version30: detected = 30; break;
+	case Version21: detected = 21; break;
+	case Version15: detected = 15; break;
+	case Version13: detected = 13; break;
+	default: break;
 	}
 
 	// Check for a requested state
 	unsigned int new_state = 0;
 	int requested = QSettings().value("GraphicsLayer", detected).toInt();
-	if (requested == 30) {
-		new_state = Version30;
-	} else if (requested == 21) {
-		new_state = Version21;
-	} else if (requested == 15) {
-		new_state = Version15;
-	} else if (requested == 13) {
-		new_state = Version13;
-	} else if (requested == 11) {
-		new_state = Version11;
-	} else {
+	QStringList args = QCoreApplication::arguments();
+	foreach (const QString& arg, args) {
+		if (arg.startsWith("--graphics-layer=")) {
+			requested = arg.mid(17).toInt();
+		}
+	}
+	switch (requested) {
+	case 30: new_state = Version30; break;
+	case 21: new_state = Version21; break;
+	case 15: new_state = Version15; break;
+	case 13: new_state = Version13; break;
+	case 11: new_state = Version11; break;
+	default:
 		new_state = state;
 		qWarning("Requested GraphicsLayer%d is invalid; using detected GraphicsLayer%d instead.", requested, detected);
+		break;
 	}
 	if (state >= new_state) {
 		state = new_state;
@@ -198,18 +251,26 @@ void GraphicsLayer::init()
 	}
 
 	// Create graphics layer instance
-	if (state == Version30) {
+	switch (state) {
+	case Version30:
+		glsl_version = (glsl <= "420") ? glsl : "420";
 		genVertexArrays(1, &vao_id);
 		bindVertexArray(vao_id);
 		graphics_layer = new GraphicsLayer21;
-	} else if (state == Version21) {
+		break;
+	case Version21:
+		glsl_version = "120";
 		graphics_layer = new GraphicsLayer21;
-	} else if (state == Version15) {
+		break;
+	case Version15:
 		graphics_layer = new GraphicsLayer15;
-	} else if (state == Version13) {
+		break;
+	case Version13:
 		graphics_layer = new GraphicsLayer13;
-	} else {
+		break;
+	default:
 		graphics_layer = new GraphicsLayer11;
+		break;
 	}
 	graphics_layer->setTextureUnits(1);
 }
@@ -373,19 +434,19 @@ GraphicsLayer21::GraphicsLayer21()
 
 	// Load shaders
 	QGLShaderProgram* program = loadProgram(0);
-	program->setAttributeBuffer(Position, GL_FLOAT, 0, 3, sizeof(Vertex));
+	program->setAttributeBuffer(Position, GL_FLOAT, offsetof(Vertex, x), 3, sizeof(Vertex));
 	program->enableAttributeArray(Position);
 
 	program = loadProgram(1);
-	program->setAttributeBuffer(TexCoord0, GL_FLOAT, sizeof(GLfloat) * 3, 2, sizeof(Vertex));
-	program->setAttributeBuffer(Position, GL_FLOAT, 0, 3, sizeof(Vertex));
+	program->setAttributeBuffer(TexCoord0, GL_FLOAT, offsetof(Vertex, s), 2, sizeof(Vertex));
+	program->setAttributeBuffer(Position, GL_FLOAT, offsetof(Vertex, x), 3, sizeof(Vertex));
 	program->enableAttributeArray(Position);
 	program->setUniformValue("texture0", GLuint(0));
 
 	program = loadProgram(2);
-	program->setAttributeBuffer(TexCoord1, GL_FLOAT, sizeof(GLfloat) * 5, 2, sizeof(Vertex));
-	program->setAttributeBuffer(TexCoord0, GL_FLOAT, sizeof(GLfloat) * 3, 2, sizeof(Vertex));
-	program->setAttributeBuffer(Position, GL_FLOAT, 0, 3, sizeof(Vertex));
+	program->setAttributeBuffer(TexCoord1, GL_FLOAT, offsetof(Vertex, s2), 2, sizeof(Vertex));
+	program->setAttributeBuffer(TexCoord0, GL_FLOAT, offsetof(Vertex, s), 2, sizeof(Vertex));
+	program->setAttributeBuffer(Position, GL_FLOAT, offsetof(Vertex, x), 3, sizeof(Vertex));
 	program->enableAttributeArray(Position);
 	program->setUniformValue("texture0", GLuint(0));
 	program->setUniformValue("texture1", GLuint(1));
@@ -470,6 +531,7 @@ void GraphicsLayer21::setTextureUnits(unsigned int units)
 
 	m_color_location = m_program->uniformLocation("color");
 	m_matrix_location = m_program->uniformLocation("matrix");
+	m_program->setUniformValue(m_color_location, Qt::white);
 	m_program->setUniformValue(m_matrix_location, m_matrix);
 
 	if (units > 1) {
@@ -512,20 +574,12 @@ QGLShaderProgram* GraphicsLayer21::loadProgram(unsigned int index)
 	}
 
 	// Update GLSL version
-	QString version;
-	version = (QGLFormat::openGLVersionFlags() & QGLFormat::OpenGL_Version_3_0) ? "130" : version;
-#if QT_VERSION >= 0x040700
-	version = (QGLFormat::openGLVersionFlags() & QGLFormat::OpenGL_Version_3_1) ? "140" : version;
-	version = (QGLFormat::openGLVersionFlags() & QGLFormat::OpenGL_Version_3_2) ? "150" : version;
-	version = (QGLFormat::openGLVersionFlags() & QGLFormat::OpenGL_Version_3_3) ? "330" : version;
-	version = (QGLFormat::openGLVersionFlags() & QGLFormat::OpenGL_Version_4_0) ? "400" : version;
-#endif
-	if (!version.isEmpty()) {
-		vertex.replace("#version 120\n", "#version " + version + "\n");
+	if (glsl_version >= "130") {
+		vertex.replace("#version 120\n", "#version " + glsl_version + "\n");
 		vertex.replace("attribute ", "in ");
 		vertex.replace("varying ", "out ");
 
-		frag.replace("#version 120\n", "#version " + version + "\n\nout vec4 out_color;\n");
+		frag.replace("#version 120\n", "#version " + glsl_version + "\n\nout vec4 out_color;\n");
 		frag.replace("gl_FragColor", "out_color");
 		frag.replace("varying ", "in ");
 	}
@@ -725,7 +779,7 @@ void GraphicsLayer15::setTextureUnits(unsigned int units)
 	if (units > 1) {
 		glEnable(GL_TEXTURE_2D);
 		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
-		glTexCoordPointer(2, GL_FLOAT, sizeof(Vertex), reinterpret_cast<GLvoid*>(sizeof(GLfloat) * 5));
+		glTexCoordPointer(2, GL_FLOAT, sizeof(Vertex), reinterpret_cast<GLvoid*>(offsetof(Vertex, s2)));
 	} else {
 		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
 		glDisable(GL_TEXTURE_2D);
@@ -736,13 +790,13 @@ void GraphicsLayer15::setTextureUnits(unsigned int units)
 	if (units > 0) {
 		glEnable(GL_TEXTURE_2D);
 		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
-		glTexCoordPointer(2, GL_FLOAT, sizeof(Vertex), reinterpret_cast<GLvoid*>(sizeof(GLfloat) * 3));
+		glTexCoordPointer(2, GL_FLOAT, sizeof(Vertex), reinterpret_cast<GLvoid*>(offsetof(Vertex, s)));
 	} else {
 		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
 		glDisable(GL_TEXTURE_2D);
 	}
 
-	glVertexPointer(3, GL_FLOAT, sizeof(Vertex), 0);
+	glVertexPointer(3, GL_FLOAT, sizeof(Vertex), reinterpret_cast<GLvoid*>(offsetof(Vertex, x)));
 }
 
 //-----------------------------------------------------------------------------
