@@ -1,6 +1,6 @@
 /***********************************************************************
  *
- * Copyright (C) 2010 Graeme Gott <graeme@gottcode.org>
+ * Copyright (C) 2010, 2011 Graeme Gott <graeme@gottcode.org>
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -30,7 +30,7 @@ class LocaleDialog : public QDialog
 public:
 	LocaleDialog(QWidget* parent = 0);
 
-	static void loadTranslator();
+	static void loadTranslator(const QString& appname);
 
 public slots:
 	virtual void accept();
@@ -43,6 +43,7 @@ private:
 
 	static QString m_current;
 	static QString m_path;
+	static QString m_appname;
 };
 
 #endif
